import logging
import pathlib

logger = logging.getLogger(__name__)


def prepare_workflow_from_template(
    workflow_template: pathlib.Path | str,
    workflow_output: pathlib.Path | str,
    database_path: pathlib.Path | str,
):
    """Prepare a FragPipe workflow file from a template by updating the database path.

    Args:
        workflow_template: Path to the template workflow file.
        workflow_output: Path to save the updated workflow file.
        database_path: Path to the FASTA database file to set in the workflow.
    """
    normalized_db_path = _resolve_path(database_path)
    updated_workflow = []
    database_path_found = False

    logger.info(f"Updating database entry in workflow template '{workflow_template}'.")

    with open(workflow_template) as template_file:
        for line in template_file.readlines():
            if line.startswith("database.db-path="):
                logger.debug(
                    f"Replacing line: {line.strip()} with "
                    f"database.db-path={normalized_db_path}"
                )

                updated_workflow.append(f"database.db-path={normalized_db_path}\n")
                database_path_found = True
            else:
                updated_workflow.append(line)

    if not database_path_found:
        logger.debug(
            "No existing database.db-path found, adding line: "
            f"database.db-path={normalized_db_path}"
        )
        updated_workflow.append(f"\ndatabase.db-path={normalized_db_path}")

    with open(workflow_output, "w") as workflow_file:
        workflow_file.write("".join(updated_workflow))


def _resolve_path(path: pathlib.Path | str) -> str:
    return pathlib.Path(path).resolve().as_posix()



import logging
import os
import pathlib
import subprocess
import time

logger = logging.getLogger(__name__)


def run_fragpipe(
    fragpipe_root: pathlib.Path | str,
    workflow_path: pathlib.Path | str,
    manifest_path: pathlib.Path | str,
    output_dir: pathlib.Path | str,
    ram: int = 0,
    threads: int = -1,
    logger: logging.Logger | None = None,
) -> bool:
    """Run FragPipe in headless mode with the specified parameters.

    If FragPipe fails to create a log file, a log file will be created manually using
    the redirected stdout.

    Tested with FragPipe v23.

    Args:
        fragpipe_root: Path to FragPipe installation directory
        workflow_path: Path to workflow file
        manifest_path: Path to manifest file
        output_dir: Path to analysis output directory
        ram: The maximum allowed memory size for FragPipe to use (in GB). Set to 0 to
            let FragPipe decide.
        threads: The number of CPU threads for FragPipe to use. Set it -1 to let
            FragPipe decide.
        logger: Logger for logging messages. If None, a default logger will be used.

    Returns:
        True if FragPipe completed successfully, False otherwise
    """

    if logger is None:
        logger = logging.getLogger(__name__)

    output_dir = pathlib.Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    redirected_log_path = output_dir / "fragpipe_stdout_redirect.log"

    if os.name == "nt":
        executable_name = "fragpipe.bat"
    elif os.name == "posix":
        executable_name = "fragpipe"
    else:
        raise OSError(f"Unsupported operating system: {os.name}")

    fragpipe_exec_path = pathlib.Path(fragpipe_root) / "bin" / executable_name
    if not fragpipe_exec_path.exists():
        raise Exception(
            f"FragPipe executable file not found at {fragpipe_exec_path}. "
            "Please check the path."
        )

    cmd = [
        fragpipe_exec_path.as_posix(),
        "--headless",
        "--workflow",
        pathlib.Path(workflow_path).resolve().as_posix(),
        "--manifest",
        pathlib.Path(manifest_path).resolve().as_posix(),
        "--workdir",
        pathlib.Path(output_dir).resolve().as_posix(),
        "--ram",
        str(ram),
        "--threads",
        str(threads),
    ]
    logger.info(f"Running FragPipe with output directory '{output_dir}'")

    try:
        start_time = time.time()
        with open(redirected_log_path, "w") as redirected_log_file:
            result = subprocess.run(
                cmd,
                check=True,
                stdout=redirected_log_file,
                stderr=subprocess.PIPE,
                text=True,
                shell=False,
            )
        duration = (time.time() - start_time) / 60
        logger.info(f"FragPipe completed successfully in {duration:.2f} minutes.")
        if result.stderr:
            logger.debug(f"FragPipe stderr output:\n{result.stderr}")

        latest_log_file = _find_latest_log_file(output_dir)
        if latest_log_file is None:
            logger.debug(
                f"No FragPipe log file found in '{output_dir}'. Using redirected log "
                "to manually create a log file."
            )
            time_stamp = time.strftime("%Y-%m-%d_%H-%M-%S")
            official_log_path = output_dir / f"log_{time_stamp}.txt"
            redirected_log_path.rename(official_log_path)
        else:
            redirected_log_path.unlink()
        return True
    except subprocess.CalledProcessError as e:
        logger.error(
            f"Error running FragPipe: {e}\nError output:\n{e.stderr}\n"
            f"A partial log file may be found at '{redirected_log_path}'"
        )
        return False


def search_results_exist(output_dir: pathlib.Path | str) -> bool:
    """Check if FragPipe search results exist in the specified output directory.

    Checks for the presence of a FragPipe log file or a combined_protein.tsv file.

    Args:
        output_dir: Path to FragPipe output directory

    Returns:
        True if search results exist, False otherwise
    """
    output_dir = pathlib.Path(output_dir)
    combined_protein_file = output_dir / "combined_protein.tsv"
    if not output_dir.exists() or not output_dir.is_dir():
        return False

    if _find_latest_log_file(output_dir) is not None:
        return True
    elif combined_protein_file.exists():
        logger.debug(
            f"FragPipe search results found in '{output_dir}', but no log file found."
        )
        return True
    else:
        return False


def clean_up_rawfile_directory(rawfile_dir: pathlib.Path):
    """Clean up FragPipe temporary files in the specified rawfile directory.

    Removes temporary files with extensions such as '.mzBIN' and '_uncalibrated.mzML'.

    Args:
        rawfile_dir: The rawfile directory to clean up.
    """
    temp_file_patterns = [
        ".mzBIN",
        "_uncalibrated.mzML",
    ]
    if not rawfile_dir.exists():
        logger.warning(f"Raw directory {rawfile_dir} does not exist.")
        return
    if not rawfile_dir.is_dir():
        logger.warning(f"Raw directory {rawfile_dir} is not a directory.")
        return

    temp_files: list[pathlib.Path] = []
    for pattern in temp_file_patterns:
        temp_files.extend(rawfile_dir.rglob(f"*{pattern}"))

    logger.debug(
        f"Trying to remove {len(temp_files)} temporary FragPipe files in {rawfile_dir}."
    )
    if not temp_files:
        return

    for temp_file in temp_files:
        temp_file.unlink()
    logger.info(f"Deleted {len(temp_files)} temporary FragPipe files in {rawfile_dir}.")


def _find_latest_log_file(output_dir: pathlib.Path) -> pathlib.Path | None:
    """Find the latest FragPipe log file in the specified output directory.

    Args:
        output_dir: Path to FragPipe output directory

    Returns:
        Path to the latest log file, or None if no log files are found
    """
    log_files = [f for f in output_dir.glob("log_*.txt") if len(f.name) == 27]
    if log_files:
        return sorted(log_files, reverse=True)[0]
    return None



import logging
import pathlib

import pandas as pd

logger = logging.getLogger(__name__)


def sdrf_to_manifest(
    sdrf_filepath: pathlib.Path | str,
    data_type: str,
    manifest_filepath: pathlib.Path | str | None = None,
    manifest_filename: str = "manifest.fp-manifest",
    experiment_field: str = "comment[data file]",
    replicate_field: str = "characteristics[biological replicate]",
) -> pd.DataFrame:
    """Convert an SDRF file to a FragPipe-compatible '.fp-manifest' file.

    Format of the tab-separated FragPipe manifest file:
    rawfile path | experiment name | bio replicate | data type

    Args:
        sdrf_filepath: Path to the SDRF file.
        data_type: Acquisition data type, one of "DDA", "DDA+", "DIA", "DIA-Quant",
            "DIA-Lib" or "GPF-DIA".
        manifest_filepath: Path to save the FragPipe manifest file. If None, saves to
            the same directory as the SDRF file.
        manifest_filename: Filename to use for the created manifest. Default is
            'manifest.fp-manifest'.
        experiment_field: Name of the SDRF column to use for the experiment information.
        replicate_field: Name of the SDRF column to use for the replicate information.

    """
    supported_data_types = ["DDA", "DDA+", "DIA", "DIA-Quant", "DIA-Lib", "GPF-DIA"]

    logger.info(f"Converting SDRF file '{sdrf_filepath}' to FragPipe manifest.")
    if data_type not in supported_data_types:
        raise ValueError(
            f"Unsupported acquisition type: {data_type}. "
            f"Supported types are: {supported_data_types}."
        )
    if manifest_filepath is None:
        manifest_filepath = pathlib.Path(sdrf_filepath).parent / manifest_filename
        logger.debug(f"No 'manifest_filepath' provided, using '{manifest_filepath}'")

    sdrf_table = pd.read_csv(sdrf_filepath, sep="\t")
    manifest = sdrf_table[
        [
            "comment[data file]",
            experiment_field,
            replicate_field,
        ]
    ].drop_duplicates()
    manifest["data type"] = data_type

    logger.info(f"Writing FragPipe manifest to '{manifest_filepath}'")
    manifest.to_csv(manifest_filepath, sep="\t", index=False, header=False)


def update_rawfile_paths_in_manifest(
    manifest_filepath: pathlib.Path | str,
    rawfile_directory: pathlib.Path | str | None = None,
) -> pd.DataFrame:
    """Update the rawfile paths in a FragPipe manifest file.

    Args:
        manifest_filepath: Path to the FragPipe manifest file.
        rawfile_directory: Directory where the rawfiles are located. If None, uses the
            directory of the manifest file.
    """
    logger.info(f"Updating rawfile paths in manifest '{manifest_filepath}'")
    if rawfile_directory is None:
        rawfile_directory = pathlib.Path(manifest_filepath).parent
        logger.debug(
            "No 'rawfile_directory' provided, using manifest "
            f"directory {rawfile_directory}"
        )
    rawfile_directory = pathlib.Path(rawfile_directory).resolve()

    manifest = pd.read_csv(manifest_filepath, sep="\t", header=None)
    rawfile_paths = [
        rawfile_directory / pathlib.Path(p).name for p in manifest.iloc[:, 0]
    ]
    manifest.iloc[:, 0] = [p.as_posix() for p in rawfile_paths]
    manifest.to_csv(manifest_filepath, sep="\t", index=False, header=False)



from .execute import run_fragpipe
from .manifest import sdrf_to_manifest, update_rawfile_paths_in_manifest
from .workflow import prepare_workflow_from_template

__all__ = [
    "prepare_workflow_from_template",
    "run_fragpipe",
    "sdrf_to_manifest",
    "update_rawfile_paths_in_manifest",
]
